from django.urls import path

import two_factor_auth.views as view


urlpatterns = [
    path(r"2fa-config/", view.Config.as_view(), name="2fa-config"),
    path(r"login/", view.UserLogin.as_view(), name="login"),
    path(r"register/", view.UserRegistrationViewSet.as_view(), name="register"),
    path(r"questions/", view.QuestionViewSet.as_view(), name="questions"),
    path(
        r"2fa-verify-answer/",
        view.VerifyAnswerViewSet.as_view(),
        name="2fa-verify-answer",
    ),
    path(r"auth-questions/", view.UserAuthAnswerView.as_view(), name="auth-questions"),
]



from hashlib import md5
from two_factor_auth.models import Question, UserAnswer


def AddQuestion(f):
    def wrapper(*args, **kwargs):
        add_question(**kwargs)

    return wrapper


def AddUserAnswer(f):
    def wrapper(*args, **kwargs):
        add_user_answer(**kwargs)

    return wrapper


def CheckAnswer(f):
    def wrapper(*args, **kwargs):
        validated = validate_answer(
            kwargs["user_id"], kwargs["question_id"], kwargs["answer"]
        )
        return f(*args, **kwargs) if validated else None

    return wrapper


def add_question(**values):
    Question.objects.create(**values)


def add_user_answer(**values):
    values["answer"] = md5(values["answer"].encode()).hexdigest()
    UserAnswer.objects.create(**values)


def validate_answer(user_id, question_id, answer):
    try:
        user_answer = UserAnswer.objects.get(user_id=user_id, question_id=question_id)
        if user_answer.answer == md5(answer.encode()).hexdigest():
            return True
        else:
            return False
    except UserAnswer.DoesNotExist:
        return False



from datetime import datetime, timedelta, timezone

from django.conf import settings
from django.db.models import F
from django.conf import settings
from rest_framework.exceptions import ValidationError

from two_factor_auth.models import TwoFactorAuthenticationSession

registration_questions_count = getattr(settings, "REGISTRATION_QUESTIONS_COUNT", 3)
registration_min_answer_count = getattr(
    settings, "USER_REGISTRATION_MIN_ANSWER_COUNT", 2
)
retry_in_seconds = getattr(
    settings, "ANSWER_ATTEMPTS_RETRY_LIFETIME", timedelta(hours=1)
)
invalid_attempt_limit = getattr(settings, "INVALID_ATTEMPTS_LIMIT", 1)


def update_2fa_session(user_id, invalid_attempt=False):
    if invalid_attempt:
        TwoFactorAuthenticationSession.objects.filter(user_id=user_id).update(
            last_answer_attempt=datetime.now(),
            invalid_answer_attempts=F("invalid_answer_attempts") + 1,
        )


def check_2fa_login_attempt(user_id):
    msg = dict()
    try:
        session, created = TwoFactorAuthenticationSession.objects.get_or_create(
            user_id=user_id,
            defaults={
                "last_answer_attempt": datetime.now(),
                "invalid_answer_attempts": 0,
            },
        )

        if session.invalid_answer_attempts >= invalid_attempt_limit:
            if (
                datetime.now(timezone.utc) - session.last_answer_attempt
                < retry_in_seconds
            ):  # noqa
                rem = (
                    session.last_answer_attempt
                    + retry_in_seconds
                    - datetime.now(timezone.utc)
                    + timedelta(minutes=0)
                )
                msg["error"] = f"Invalid Attempts Exceeded, Try After {rem}"
                raise ValidationError(msg, code="authorization")
            else:
                session.last_answer_attempt = datetime.now()
                session.invalid_answer_attempts = 0
                session.save()

    except TwoFactorAuthenticationSession.DoesNotExist:
        msg["error"] = "Invalid User"
        raise ValidationError(msg, code="authorization")



from django.contrib.auth import get_user_model
from django.db.models import (
    CASCADE,
    IntegerField,
    DateTimeField,
    ForeignKey,
    Model,
    PositiveSmallIntegerField,
    OneToOneField,
    TextField,
    UUIDField,
)
from orm_choices import choices
from uuid import uuid4

User = get_user_model()


@choices
class AnswerType:
    # TODO : add extension on backend to add choices
    class Meta:
        TEXT = (1, "Text")
        SELECT = (2, "Select")


class BaseModel(Model):
    id = UUIDField(default=uuid4, primary_key=True, editable=False)
    created_at = DateTimeField(auto_now_add=True, editable=False)
    updated_at = DateTimeField(auto_now=True, editable=False)

    class Meta:
        abstract = True


class TwoFactorAuthenticationSession(BaseModel):
    user = OneToOneField(User, on_delete=CASCADE)
    last_answer_attempt = DateTimeField(null=True)
    invalid_answer_attempts = IntegerField()

    class Meta:
        verbose_name = "2FA Session"
        verbose_name_plural = "2FA Sessions"

    def __str__(self):
        return str(self.id)


class Question(BaseModel):
    question_desc = TextField()
    answer_type = PositiveSmallIntegerField(
        default=AnswerType.TEXT, choices=AnswerType.CHOICES
    )

    class Meta:
        verbose_name = "Question"
        verbose_name_plural = "Questions"

    def __str__(self):
        return self.question_desc


class UserAnswer(BaseModel):
    django_user = ForeignKey(User, on_delete=CASCADE)
    question = ForeignKey(Question, on_delete=CASCADE)
    answer = TextField()

    class Meta:
        verbose_name = "User Answer"
        verbose_name_plural = "User Answers"
        unique_together = ["django_user", "question"]

    def __str__(self):
        return str(self.id)



from hashlib import md5

from django.contrib.auth import get_user_model, authenticate
from rest_framework.authtoken.models import Token
from rest_framework.serializers import ModelSerializer
from rest_framework.serializers import (
    Serializer,
    CharField,
    ValidationError,
    SerializerMethodField,
    UUIDField,
)

from two_factor_auth.models import Question, UserAnswer
from two_factor_auth.models import User

User = get_user_model()


class UserAuthInputSerializer(Serializer):
    username = CharField()
    password = CharField()

    def validate(self, attrs):
        username = attrs.get("username")
        password = attrs.get("password")

        if username and password:
            user = authenticate(username=username, password=password)
            if user:

                if not user.is_active:
                    msg = "User account is disabled."
                    raise ValidationError(msg, code="authorization")

            else:
                msg = "Unable to log in with provided credentials."
                raise ValidationError(msg, code="authorization")

        else:
            msg = 'Must include "username" and "password".'
            raise ValidationError(msg, code="authorization")

        attrs["user"] = user
        return attrs


class QuestionSerializer(ModelSerializer):
    class Meta:
        model = Question
        fields = "__all__"


class UserQuestionSerializer(ModelSerializer):
    id = SerializerMethodField()
    question = SerializerMethodField()

    class Meta:
        model = UserAnswer
        fields = ("id", "question")

    def get_id(self, obj):
        return obj.question.id

    def get_question(self, obj):
        return obj.question.question_desc


class LoginSerializer(ModelSerializer):
    questions = SerializerMethodField()

    class Meta:
        model = User
        fields = (
            "username",
            "first_name",
            "last_name",
            "email",
            "questions",
            "auth_token",
        )

    def get_questions(self, obj):
        return UserQuestionSerializer(obj.useranswer_set.filter(), many=True).data


class QuestionSerializer(ModelSerializer):
    class Meta:
        model = Question
        fields = "__all__"


class CreateUserAnswerSerializer(ModelSerializer):
    question_id = UUIDField(required=True)

    class Meta:
        model = UserAnswer
        fields = ("question_id", "answer")

    def create(self, validated_data):
        django_user = self.context.get("django_user")
        validated_data["django_user"] = django_user
        validated_data["answer"] = md5(validated_data["answer"].encode()).hexdigest()
        return UserAnswer.objects.create(**validated_data)


class UserSerializer(ModelSerializer):
    class Meta:
        model = User
        fields = "__all__"


class CreateUserSerializer(ModelSerializer):
    password = CharField(write_only=True)
    auth_token = CharField(required=False)

    class Meta:
        model = User
        fields = ("first_name", "last_name", "email", "password", "auth_token")

    def create(self, validated_data):
        validated_data["username"] = validated_data["email"]
        user = User.objects.create_user(**validated_data)
        token, created = Token.objects.get_or_create(user=user)
        user.auth_token = token
        user.save()
        return user


class AnswerRequestInputSerializer(Serializer):
    question_id = CharField()
    answer = CharField()


class VerifyAnswerInputSerializer(Serializer):
    requests = AnswerRequestInputSerializer(many=True)


class VerifyAnswerSerializer(Serializer):
    refresh = CharField()
    access = CharField()



# Generated by Django 3.2.4 on 2021-06-27 03:20

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Question",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("question_desc", models.TextField()),
                (
                    "answer_type",
                    models.PositiveSmallIntegerField(
                        choices=[(2, "Select"), (1, "Text")], default=1
                    ),
                ),
            ],
            options={
                "verbose_name": "Question",
                "verbose_name_plural": "Questions",
            },
        ),
        migrations.CreateModel(
            name="TwoFactorAuthenticationSession",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("last_answer_attempt", models.DateTimeField(null=True)),
                ("invalid_answer_attempts", models.IntegerField()),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "2FA Session",
                "verbose_name_plural": "2FA Sessions",
            },
        ),
        migrations.CreateModel(
            name="UserAnswer",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("answer", models.TextField()),
                (
                    "django_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "question",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="two_factor_auth.question",
                    ),
                ),
            ],
            options={
                "verbose_name": "User Answer",
                "verbose_name_plural": "User Answers",
                "unique_together": {("django_user", "question")},
            },
        ),
    ]



from django.contrib import admin

from two_factor_auth.models import Question, TwoFactorAuthenticationSession, UserAnswer

admin.site.register(Question)
admin.site.register(TwoFactorAuthenticationSession)
admin.site.register(UserAnswer)



from hashlib import md5

from django.contrib.auth import get_user_model
from django.db import transaction
from rest_framework.authentication import TokenAuthentication
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.generics import ListAPIView, CreateAPIView
from rest_framework.permissions import AllowAny
from rest_framework import status
from rest_framework.views import APIView
from rest_framework_simplejwt.tokens import RefreshToken

from two_factor_auth.serializers import CreateUserSerializer, CreateUserAnswerSerializer
from two_factor_auth.models import Question, UserAnswer
from two_factor_auth.serializers import (
    QuestionSerializer,
    LoginSerializer,
    UserAuthInputSerializer,
    VerifyAnswerInputSerializer,
    VerifyAnswerSerializer,
)
from two_factor_auth.utils import (
    check_2fa_login_attempt,
    update_2fa_session,
    invalid_attempt_limit,
    registration_questions_count,
    registration_min_answer_count,
    retry_in_seconds,
)

User = get_user_model()


class UserLogin(GenericAPIView):
    serializer_class = LoginSerializer
    permission_classes = (AllowAny,)

    def post(self, request):
        serializer = UserAuthInputSerializer(data=request.data)
        if serializer.is_valid():
            user = serializer.validated_data["user"]
            return Response(
                {"user": self.get_serializer(user).data}, status=status.HTTP_200_OK
            )
        return Response(serializer.errors, status=status.HTTP_401_UNAUTHORIZED)


class QuestionViewSet(ListAPIView):
    queryset = Question.objects.all()
    serializer_class = QuestionSerializer
    permission_classes = (AllowAny,)


class VerifyAnswerViewSet(GenericAPIView):
    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)
    serializer_class = VerifyAnswerSerializer

    def post(self, request):
        serializer = VerifyAnswerInputSerializer(data=request.data)
        if serializer.is_valid():
            answer_count_temp = 0
            user_id = request.user.id

            check_2fa_login_attempt(user_id)

            for rq in request.data.get("requests"):
                question_id = rq.get("question_id")
                answer_encode = md5(rq.get("answer").encode()).hexdigest()

                answer = UserAnswer.objects.filter(
                    django_user_id=user_id,
                    question_id=question_id,
                    answer__exact=answer_encode,
                )

                if answer.count() > 0:
                    answer_count_temp += 1

            if answer_count_temp >= invalid_attempt_limit:
                data = dict()
                user = User.objects.get(id=user_id)
                refresh = RefreshToken.for_user(user)
                data["refresh"] = str(refresh)
                data["access"] = str(refresh.access_token)
                return Response(
                    {"token": self.get_serializer(data).data}, status=status.HTTP_200_OK
                )
            else:
                update_2fa_session(user_id, True)
                msg = dict()
                msg["error"] = "Invalid Answers"
                return Response(msg, status=status.HTTP_401_UNAUTHORIZED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class UserRegistrationViewSet(CreateAPIView):
    queryset = User.objects.all()
    serializer_class = CreateUserSerializer
    permission_classes = (AllowAny,)


class UserAuthAnswerView(GenericAPIView):
    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)

    def post(self, request):
        data = request.data
        requests_data = data.get("requests")
        try:
            user_obj = User.objects.get(id=request.user.id, is_active=True)
        except User.DoesNotExist:
            return Response("User does not exist")
        with transaction.atomic():
            user_answer_serializer = CreateUserAnswerSerializer(
                data=requests_data, many=True, context={"django_user": user_obj}
            )
            if not user_answer_serializer.is_valid():
                transaction.set_rollback(True)
                return Response(
                    user_answer_serializer.errors, status=status.HTTP_400_BAD_REQUEST
                )
            user_answer_serializer.save()

        return Response("success", status=status.HTTP_200_OK)


class Config(APIView):
    permission_classes = (AllowAny,)

    def get(self, request):
        return Response(
            {
                "Config": {
                    "registration_questions_count": registration_questions_count,
                    "registration_min_answer_count": registration_min_answer_count,
                    "retry_in_seconds": retry_in_seconds,
                    "invalid_attempt_limit": invalid_attempt_limit,
                }
            }
        )



from distutils.core import setup

setup(
    name="django-two-factor-auth-qna",  # How you named your package folder (MyLib)
    packages=[
        "two_factor_auth",
        "two_factor_auth/migrations",
    ],  # Chose the same as "name"
    version="0.0.8",  # Start with a small number and increase it with every change you make
    license="MIT",  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description="Package to enable two factor authentication using Questions and answers",  # Give a short description about your library
    author="TeamX",  # Type in your name
    author_email="teamx@reckonsys.com",  # Type in your E-Mail
    url="https://github.com/rsys-teamx/django-2fa-qna",  # Provide either the link to your github or to your website
    download_url="https://github.com/rsys-teamx/django-2fa-qna/",
    keywords=["2FA", "Reckonsys Hackathon Submission", "team"],
    install_requires=[
        "jwt",
        "Django",
        "djangorestframework",
        "orm_choices",
        "djangorestframework-simplejwt",
        "djangorestframework-camel-case"
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Developers",  # Define that your audience are developers
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",  # Again, pick a license
        "Programming Language :: Python :: 3.9"
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.4',
        # 'Programming Language :: Python :: 3.5',
        # 'Programming Language :: Python :: 3.6',
    ],
)
